#include "computation.h"

bool dbl(natural_t *nat) {
 xit_carry_t carry = 0;
 for (size_t j = 0; j < nat->n_xits; j++)
  nat->xits[j] = (carry += 2*nat->xits[j]) % nat->base, carry /= nat->base;
 if (carry) {
  if (nat->n_xits+1 >= nat->n_xits_max)
   return nat->overflowed = true, ERROR("Overflow."), false;
  nat->xits[nat->n_xits++] = carry; }
 return true; }

bool inc(natural_t *nat, size_t p) {
 for (size_t j = p; j < nat->n_xits; j++)
  if (nat->xits[j]+1 == nat->base) nat->xits[j] = 0;
  else return nat->xits[j]++, true;
 nat->n_xits++;
 if (nat->n_xits >= nat->n_xits_max)
  return nat->overflowed = true, ERROR("Overflow."), false;
 return nat->xits[nat->n_xits-1] = 1, true; }

bool hlv(natural_t *nat) {
 bool frac = false;
 for (size_t j = nat->n_xits-1; j != -1; j--) {
  bool odd = nat->xits[j] % 2;
  nat->xits[j] = nat->xits[j] / 2 + nat->base/2*frac;
  if (nat->base % 2 == 0) frac = odd;
  else if (odd) frac = frac ? inc(nat, j), false : true; }
 if (nat->n_xits >= 1 && nat->xits[nat->n_xits-1] == 0)
  nat->n_xits--;
 return !frac; }


rit_genus_t rit_genus(natural_t *nat) {
 rit_count_t r0 = 0, r1 = 0, r2 = 0;
 for (size_t j = 0; j < nat->n_xits; j++) {
  xit_t x = nat->xits[j]; r0 += xr0map[x], r1 += xr1map[x], r2 += xr2map[x]; }
 if (nat->n_xits >= 1) r0 -= xrl0map[nat->xits[nat->n_xits-1]];
 return (rit_genus_t) { .r0 = r0, .r1 = r1, .r2 = r2 }; }

bool rit_geni_equal(rit_genus_t g, rit_genus_t h) {
 return g.r0 == h.r0 && g.r1 == h.r1 && g.r2 == h.r2; }

bool dbl_rit(natural_t *nat, rit_genus_t *gen) {
 gen->r0 = gen->r1 = gen->r2 = 0;
 xit_carry_t carry = 0;
 for (size_t j = 0; j < nat->n_xits; j++) {
  xit_t x = (carry += 2*nat->xits[j]) % nat->base;
  carry /= nat->base;
  nat->xits[j] = x; gen->r0 += xr0map[x];
  gen->r1 += xr1map[x], gen->r2 += xr2map[x]; }
 if (carry) {
  if (nat->n_xits+1 >= nat->n_xits_max)
   return nat->overflowed = true, ERROR("Overflow."), false;
  nat->xits[nat->n_xits++] = carry; gen->r0 += xr0map[carry];
  gen->r1 += xr1map[carry], gen->r2 += xr2map[carry]; }
 if (nat->n_xits >= 1) gen->r0 -= xrl0map[nat->xits[nat->n_xits-1]];
 return true; }


void free_xit_rit_mapping() {
 free(xrl0map), free(xr0map), free(xr1map), free(xr2map); }

bool build_xit_rit_mapping(computation_t *computation) {
 natural_t *nat = &computation->two_pow_ns[computation->n_terms-1];
 xit_t radix = nat->radix;
 xit_t base = nat->base;

 LOG_INIT, LOG_INTR("Allocating xit-rit mapping (");
 LOG_INTR_WITH(fprint_byte_size, 4*base * sizeof *xr0map);
 LOG_INTR(") ..."), LOG_FNAL;
 clock_t allocation_start = clock();
 xrl0map = malloc(base * sizeof *xrl0map);
 xr0map = malloc(base * sizeof *xr0map);
 xr1map = malloc(base * sizeof *xr1map);
 xr2map = malloc(base * sizeof *xr2map);
 if (!xrl0map || !xr0map || !xr1map || !xr2map)
  return free_xit_rit_mapping(), MALLOC_ERROR(), false;
 LOG("Allocated xit-rit mapping in %.4f seconds."
    , (double) (clock() - allocation_start) / CLOCKS_PER_SEC);

 LOG("Building xit-rit mapping ...");
 for (xit_t _x = 0; _x < base; _x++) {
  xrl0map[_x] = xr0map[_x] = xr1map[_x] = xr2map[_x] = 0;
  xit_t x = _x;
  for (xit_t rpow = base; rpow /= radix; x %= rpow)
   switch (x / rpow) {
    case 0: xrl0map[_x] += !xr1map[_x] && !xr2map[_x];
            xr0map[_x]++; break;
    case 1: xr1map[_x]++; break;
    case 2: xr2map[_x]++; break; }}
 return LOG("Built xit-rit mapping."), true; }

bool new_computation(pargs_t *pargs, computation_t *computation) {
 uint64_t _base = intpow(radix, pargs->three_pow);
 if (_base != (xit_t) _base)
  return ERROR("Proposed base %" PRIu64 " exceeds xit_t's bounds (2^(8*%zu)-1 "
               "< base).", _base, sizeof (xit_t)), false;
 if (_base*2-1 != (xit_carry_t) (_base*2-1))
  return ERROR("Proposed base %" PRIu64 " exceeds xit_carry_t's bounds "
               "(2^(8*%zu)-1 < 2*base-1).", _base, sizeof (xit_carry_t))
         , false;
 xit_t base = (xit_t) _base;
 LOG("Set base as %" PRIu64 "^%" PRIu64 " = %" PRIu64 ".", (uint64_t) radix
    , pargs->three_pow, _base);

 size_t n_xits = 1;
 xit_t *xits = calloc(n_xits_max, sizeof (xit_t));
 if (!xits) return MALLOC_ERROR(), false;
 xits[0] = 1;

 natural_t two_pow_n = (natural_t) {
  .radix = radix, .base = base,
  .n_xits_max = n_xits_max, .n_xits = n_xits, .xits = xits,
  .overflowed = false, .sequence_member = false,
  .time = 0 };

 size_t n_terms = 1;
 uint64_t *ns = malloc(n_terms * sizeof *ns);
 if (!ns) return free(xits), MALLOC_ERROR(), false;

 natural_t *two_pow_ns = malloc(n_terms * sizeof *two_pow_ns);
 if (!two_pow_ns) return free(xits), free(ns), MALLOC_ERROR(), false;
 ns[0] = 0;
 two_pow_ns[0] = two_pow_n;

 size_t comment_length = str_len(pargs->comment);
 char *comment = malloc((comment_length+1) * sizeof *comment);
 if (!comment) return free(ns), free(two_pow_ns), MALLOC_ERROR(), false;
 for (size_t j = 0; j < comment_length; j++)
  comment[j] = pargs->comment[j];
 comment[comment_length] = '\0';

 computation->n_terms = n_terms;
 computation->ns = ns;
 computation->two_pow_ns = two_pow_ns;
 computation->comment = comment;
 return true; }

void free_computation(computation_t *computation) {
 free(computation->comment);
 if (computation->two_pow_ns)
  for (size_t j = 0; j < computation->n_terms; j++)
   free(computation->two_pow_ns[j].xits);
 free(computation->two_pow_ns);
 free(computation->ns); }


static bool extend_two_pow_ns(
 uint64_t *n, natural_t *two_pow_n, bool sequence_member,
 computation_t *computation
) {
 LOG("Extending file with %" PRIu64 " ...", *n);
 if (computation->n_terms < 1) return ERROR("Too few terms."), false;

 size_t n_terms = computation->n_terms + 1;
 uint64_t *ns = realloc(computation->ns, n_terms * sizeof *ns);
 if (!ns) return REALLOC_ERROR(), false;
 computation->ns = ns;
 n = &computation->ns[n_terms-2];

 natural_t *two_pow_ns = realloc(computation->two_pow_ns
                                , n_terms * sizeof *two_pow_ns);
 if (!two_pow_ns) return REALLOC_ERROR(), false;
 computation->two_pow_ns = two_pow_ns;
 two_pow_n = &computation->two_pow_ns[n_terms-2];

 computation->ns[n_terms-1] = *n;
 computation->two_pow_ns[n_terms-1] = *two_pow_n;
 xit_t *xits = malloc(two_pow_n->n_xits * sizeof *xits);
 if (!xits) return MALLOC_ERROR(), false;
 for (size_t j = 0; j < two_pow_n->n_xits; j++)
  xits[j] = two_pow_n->xits[j];

 computation->two_pow_ns[n_terms-2] = (natural_t) {
  .radix = two_pow_n->radix, .base = two_pow_n->base,
  .n_xits = two_pow_n->n_xits, .n_xits_max = two_pow_n->n_xits /* on purpose */,
  .xits = xits, .overflowed = two_pow_n->overflowed,
  .sequence_member = sequence_member,
  .time = computation->two_pow_ns[n_terms-1].time };

 return computation->n_terms = n_terms, true; }

bool compute_computation(pargs_t *pargs, computation_t *computation) {
 LOG("Starting computation ...");
 clock_t computation_time = clock();

 uint64_t total_computation_seconds
  = computation->two_pow_ns[computation->n_terms-1].time;
 uint64_t computation_seconds_of_last_save = 0
        , computation_seconds_since_last_save = 0;
 #define COMPUTATION_SECONDS \
  (total_computation_seconds + (clock() - computation_time) / CLOCKS_PER_SEC)

 #define SAVE_OVERDUE \
  (computation_seconds_since_last_save >= save_frequency_seconds)

 #define UPDATE_TIME \
  computation->two_pow_ns[computation->n_terms-1].time = COMPUTATION_SECONDS

 #define EXTEND(SEQUENCE_MEMBER) \
  UPDATE_TIME; \
  if (!extend_two_pow_ns(n, two_pow_n, SEQUENCE_MEMBER, computation)) \
   return ERROR("Could not extend."), false; \
  n = &computation->ns[computation->n_terms-1]; \
  two_pow_n = &computation->two_pow_ns[computation->n_terms-1];

 #define SAVE \
  if (!pargs->nosave && !write_computation(pargs->filename, computation)) \
   FILE_WRITE_ERROR(pargs->filename);

 uint64_t *n = &computation->ns[computation->n_terms-1];
 natural_t *two_pow_n = &computation->two_pow_ns[computation->n_terms-1];

 if (two_pow_n->n_xits_max != n_xits_max)
  return ERROR("File's n_xits_max = %zu != %zu differs.", two_pow_n->n_xits_max
              , n_xits_max), false;

 uint64_t initial_n = *n;
 uint64_t target = pargs->target;
 if (pargs->target_plus) {
  if (initial_n > UINT64_MAX-target)
   return ERROR("Target addition exceeds value bounds."), false;
  target += initial_n;
  LOG("n = %" PRIu64 ", target = %" PRIu64 "+%" PRIu64 " = %" PRIu64 ""
     , initial_n, initial_n, pargs->target, target); }
 else {
  LOG("n = %" PRIu64 ", target = %" PRIu64 "", initial_n, target); }
 if (initial_n >= target)
  return LOG("Target already reached."), true;

 size_t log2_max_rit_count = (size_t) ceil(log2(target/log2(two_pow_n->radix)));
 if (log2_max_rit_count >= 8 * sizeof (rit_count_t))
  return ERROR("Approximation of maximum rit count 2^%zu exceeds rit_count_t's "
               "bounds (2^(8*%zu)-1).", log2_max_rit_count
               , sizeof (rit_count_t)), false;
 if (pargs->only_verify_rit_count_t)
  return true;

 rit_genus_t genus_two_pow_n, genus_two_pow_np1 = rit_genus(two_pow_n);
 bool first_iteration = true;
 for (; *n < target; ++*n) {
  if (!first_iteration && !pargs->nosave && *n) {
   if (*n % save_frequency_seconds_update_frequency_n == 0)
    computation_seconds_since_last_save = COMPUTATION_SECONDS
                                          - computation_seconds_of_last_save;
   if (*n % save_frequency_n == 0 || SAVE_OVERDUE) {
    EXTEND(false);
    SAVE;
    if (SAVE_OVERDUE) {
     LOG("Computed for %.4f seconds.", (double) (clock() - computation_time)
                                                / CLOCKS_PER_SEC);
     computation_seconds_of_last_save = COMPUTATION_SECONDS;
     computation_seconds_since_last_save = 0; }
  }}

  genus_two_pow_n = genus_two_pow_np1;
  bool successful;
  if (pargs->one_pass) successful = dbl_rit(two_pow_n, &genus_two_pow_np1);
  else successful = dbl(two_pow_n), genus_two_pow_np1 = rit_genus(two_pow_n);
  if (!successful)
   return ERROR("Failed to double 2^%" PRIu64 ".", *n), false;

  if (rit_geni_equal(genus_two_pow_n, genus_two_pow_np1)) {
   LOG("@@@ Found new sequence member %" PRIu64 ". @@@", *n);
   if (!pargs->nosave) {
    hlv(two_pow_n);
    EXTEND(true);
    dbl(two_pow_n);
    SAVE; } }

 first_iteration = false; }

 UPDATE_TIME; SAVE;
 uint64_t final_n = computation->ns[computation->n_terms-1];

 double computation_seconds = (double) (clock() - computation_time)
                                       / CLOCKS_PER_SEC;
 LOG("Ended computation from n=%" PRIu64 " to n=%" PRIu64 ". Took %.4f seconds."
    , initial_n, final_n, computation_seconds);

 if (pargs->print_time) {
  printf("# computation_seconds (sizeof (xit_t)) (sizeof (xit_carry_t)) "
         "(sizeof (rit_count_t)) initial_n final_n\n");
  printf("%.4f %zu %zu %zu %" PRIu64 " %" PRIu64 "\n", computation_seconds
        , sizeof (xit_t), sizeof (xit_carry_t), sizeof (rit_count_t)
        , initial_n, final_n); }

 #undef COMPUTATION_SECONDS
 #undef SAVE_OVERDUE
 #undef UPDATE_TIME
 #undef EXTEND
 #undef SAVE
 return true; }

bool edit_computation(pargs_t *pargs, computation_t *computation) {
 uint64_t removed = 0, kept = 0;
 if (pargs->remove_non_members) {
  for (size_t j = 0; j+1 < computation->n_terms; j++) {
   if (!computation->two_pow_ns[j].sequence_member) {
    LOG("Removing %" PRIu64 " ...", computation->ns[j]);
    free(computation->two_pow_ns[j].xits);
    for (size_t i = j; i+1 < computation->n_terms; i++) {
     computation->ns[i] = computation->ns[i+1];
     computation->two_pow_ns[i] = computation->two_pow_ns[i+1]; }
    computation->n_terms--; j--;
    removed++; }
   else LOG("Kept %" PRIu64 ".", computation->ns[j]), kept++; }
  LOG("In total: removed %" PRIu64 " element%s, kept %" PRIu64 ". Now %"
      PRIu64 " element%s.", removed , removed == 1 ? "" : "s", kept
      , (uint64_t) computation->n_terms, computation->n_terms == 1 ? "" : "s");
 }

 if (pargs->new_comment) {
  char *cmt = malloc((str_len(pargs->new_comment)+1) * sizeof *cmt);
  if (!cmt) return MALLOC_ERROR(), false;
  for (size_t j = 0; pargs->new_comment[j]; j++)
   cmt[j] = pargs->new_comment[j];
  cmt[str_len(pargs->new_comment)] = '\0';
  free(computation->comment); computation->comment = cmt; }

 if (!write_computation(pargs->filename, computation))
  return FILE_WRITE_ERROR(pargs->filename), false;
 return true; }


bool fprint_as_ternary(
 FILE *f, bool plus_one, uint64_t n, computation_t *computation
) {
 uint64_t n_under = 0, n_over = UINT64_MAX;
 bool found_n_under = false, found_n_over = false;
 for (size_t j = 0; j < computation->n_terms; j++) {
  if (n_under < computation->ns[j] && computation->ns[j] < n)
   found_n_under = true, n_under = computation->ns[j];
  if (computation->ns[j] == n) {
   if (plus_one) dbl(&computation->two_pow_ns[j]);
   return fprint_rits(f, &computation->two_pow_ns[j]), fprintf(f, "\n"), true; }
  if (n < computation->ns[j] && computation->ns[j] < n_over)
   found_n_over = true, n_over = computation->ns[j]; }
 ERROR("Could not find %" PRIu64 ".", n);
 if (found_n_under)
  ERROR("Closest found entry under %" PRIu64 " is %" PRIu64 ".", n, n_under);
 if (found_n_over)
  ERROR("Closest found entry over %" PRIu64 " is %" PRIu64 ".", n, n_over);
 return false; }

int fprint_rits(FILE *f, natural_t *nat) {
 int chars = 0; bool non_zero = false;
 for (size_t j = nat->n_xits-1; j != -1; j--) {
  xit_t x = nat->xits[j];
  for (xit_t rpow = nat->base; rpow /= nat->radix; x %= rpow) {
   xit_t rit = x / rpow;
   if ((non_zero |= !!rit)) {
    /**/ if (0 <= rit && rit <= 9) putc('0' + rit, f), chars++;
    else if (10 <= rit && rit <= 36) putc('A' + (rit-10), f), chars++;
    else chars += fprintf(f, "{%" PRIu64 "}", (uint64_t) rit); }}}
 return chars + fprintf(f, "%.*s", !non_zero, "0"); }

int fprint_rits_reversed(FILE *f, natural_t *nat) {
 int chars = 0;
 for (size_t j = 0; j < nat->n_xits; j++) {
  xit_t x = nat->xits[j];
  for (xit_t rpow = nat->base; rpow /= nat->radix; x /= nat->radix) {
   xit_t rit = x % nat->radix;
   /**/ if (0 <= rit && rit <= 9) putc('0' + rit, f), chars++;
   else if (10 <= rit && rit <= 36) putc('A' + (rit-10), f), chars++;
   else chars += fprintf(f, "{%" PRIu64 "}", (uint64_t) rit); }}
 return chars; }

int fprint_rits_as_expression(FILE *f, natural_t *nat) {
 int chars = 0; uint64_t total_pow = 0;
 for (size_t j = 0; j < nat->n_xits; j++) {
  xit_t x = nat->xits[j];
  for (xit_t rpow = nat->base; rpow /= nat->radix; x /= nat->radix) {
   xit_t rit = x % nat->radix;
   /**/ if (rit == 0) ;
   else if (rit == 1) chars += fprintf(f, "(%" PRIu64 "**%" PRIu64 ")+"
                                      , (uint64_t) nat->radix, total_pow);
   else chars += fprintf(f, "(%" PRIu64 "*%" PRIu64 "**%" PRIu64 ")+"
                        , (uint64_t) rit, (uint64_t) nat->radix, total_pow);
   total_pow++; }}
 return chars + fprintf(f, "0"); }

int fprint_natural(FILE *f, natural_t *nat) {
 int chars = 0;
 if (nat->overflowed) chars += fprintf(f, "(OVERFLOWED) ");
 chars += fprintf(f, "Represented in base %" PRIu64 "^%" PRIu64 "=%" PRIu64 ", "
                     "using ", (uint64_t) nat->radix
                 , intceillog(nat->radix, nat->base), (uint64_t) nat->base);
 chars += fprint_byte_size(f, nat->n_xits)
        + fprintf(f, "/")
        + fprint_byte_size(f, nat->n_xits_max);
 if (nat->n_xits_max != 0)
  chars += fprintf(f, " (%.2f%%)", 100 * (double) nat->n_xits
                                   / nat->n_xits_max);
 return chars + printf(". Computed in %" PRIu64 " second%s.", nat->time
                      , nat->time == 1 ? "" : "s"); }

int fprint_computation(FILE *f, char *filename, computation_t *computation) {
 int chars = 0;
 chars += fprintf(f, "FILE \"") + fprint_escaped(f, filename)
        + fprintf(f, "\"\n\n") + fprintf(f, "  COMMENT\n");
 if (*(computation->comment))
  chars += fprintf(f, "    \"")
         + fprint_escaped(f, computation->comment)
         + fprintf(f, "\"\n");
 else chars += fprintf(f, "    (no comment)\n");

 chars += fprintf(f, "\n")
        + fprintf(f, "  DATATYPE SIZES\n")
        + fprintf(f, "    sizeof (xit_t)       = %zu\n", sizeof (xit_t))
        + fprintf(f, "    sizeof (xit_carry_t) = %zu\n", sizeof (xit_carry_t))
        + fprintf(f, "    sizeof (rit_count_t) = %zu\n", sizeof (rit_count_t));

 #define PRINT_NATURAL(J) \
   chars += fprintf(f, "    2^%" PRIu64 ": ", computation->ns[J]) \
          + fprint_natural(f, &computation->two_pow_ns[J]) \
          + fprintf(f, "\n")

 #define SEQ(NAME, LOGIC) { \
  chars += fprintf(f, "\n  " NAME "\n"); \
  bool any = false; \
  for (size_t j = 0; j+1 < computation->n_terms; j++) \
   if (LOGIC computation->two_pow_ns[j].sequence_member) { \
    PRINT_NATURAL(j); \
    any = true; } \
  if (!any) \
   chars += fprintf(f, "    (none yet computed)\n"); }

 SEQ("NON-SEQUENCE TERMS", !)
 SEQ("SEQUENCE TERMS", !!)
 chars += fprintf(f, "\n  ACTIVE\n");
 PRINT_NATURAL(computation->n_terms-1);

 #undef PRINT_NATURAL
 #undef SEQ
 return chars; }



