#ifndef DATATYPES_H
#define DATATYPES_H
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>

/* Needs to hold values in {0, ..., base-1, base}. */
#ifndef XIT_T
#define XIT_T uint32_t
#endif
typedef XIT_T xit_t;

/* Needs to hold values in {0, ..., 2*(base-1)+1 = 2*base-1}. */
#ifndef XIT_CARRY_T
#define XIT_CARRY_T uint32_t
#endif
typedef XIT_CARRY_T xit_carry_t;

/* Needs to hold values in {0, ..., ceil(log2(target / log2(radix)))}. */
#ifndef RIT_COUNT_T
#define RIT_COUNT_T uint32_t
#endif
typedef RIT_COUNT_T rit_count_t;

typedef struct {
 char *filename;
 bool log;

 bool mode_help;

 bool mode_new;
 uint64_t three_pow;
 char *comment;

 bool mode_info;
 uint64_t as_ternary;
 bool as_ternary_plus_one;

 bool mode_compute;
 uint64_t target;
 bool target_plus;
 bool nosave;
 bool print_time;
 bool only_verify_rit_count_t;
 bool one_pass;

 bool mode_edit;
 bool remove_non_members;
 char *new_comment;
} pargs_t;


typedef struct {
 xit_t radix;
 xit_t base;

 /* Should adhere to `1 <= n_xits <= n_xits_max`. */
 size_t n_xits_max;
 size_t n_xits;
 xit_t *xits;

 bool overflowed;
 bool sequence_member;

/* In seconds; represents computational time attached to this natural. */
 uint64_t time;
} natural_t;


typedef struct {
 char *comment;

 size_t n_terms;
 uint64_t *ns;
 natural_t *two_pow_ns;

} computation_t;


typedef struct {
 rit_count_t r0, r1, r2;
} rit_genus_t;

#endif
