/* Jonathan Frech, 17th, 18th, 19th, 20th, 22nd, 23rd and 25th, 26th, 27th,
                   28th, 29th of September, 1st, 2nd, 3rd, 4th, 10th, 14th,
                   15th, 18th, 20th of October 2019 */
/* Generates terms in OEIS sequence A056154. */

#include "main.h"

rit_count_t *xrl0map = NULL, *xr0map = NULL, *xr1map = NULL, *xr2map = NULL;
bool do_log = false;
uint64_t save_frequency_n = 1000000, save_frequency_seconds = 60*60*8,
         save_frequency_seconds_update_frequency_n = 100000;
const xit_t radix = 3;
const uint64_t max_target = 1000000000;
const size_t n_xits_max = 0xFFFFFF;
const uint64_t default_three_pow = 11;

int main(int argc, char **argv) {
 pargs_t pargs;
 if (!parse_args(argc, argv, &pargs)) return EXIT_FAILURE;
 do_log = pargs.log;
 if (pargs.mode_help) return print_help(), EXIT_SUCCESS;

 bool successful = true;
 computation_t computation = (computation_t) {
  .comment = NULL, .n_terms = 0, .ns = NULL, .two_pow_ns = NULL };

 #define FAULTY_FILE \
  free_computation(&computation), \
  ERROR("File '%s' could not be read or contains faulty data." \
       , pargs.filename), EXIT_FAILURE

 if (pargs.mode_new) {
  if (!access(pargs.filename, F_OK))
   return ERROR("File '%s' already exists.", pargs.filename), EXIT_FAILURE;
  successful = new_computation(&pargs, &computation)
             && build_xit_rit_mapping(&computation)
             && write_computation(pargs.filename, &computation); }

 else if (pargs.mode_info) {
  if (!read_computation(pargs.filename, &computation)) { return FAULTY_FILE; }
  if (!build_xit_rit_mapping(&computation)) {
   free_computation(&computation);
   return ERROR("Could not build xit-rit mapping."), EXIT_FAILURE; }
  if (pargs.as_ternary)
   successful = fprint_as_ternary(stdout, pargs.as_ternary_plus_one
                                 , pargs.as_ternary, &computation);
  else fprint_computation(stdout, pargs.filename, &computation); }

 else if (pargs.mode_compute) {
  if (!read_computation(pargs.filename, &computation)) { return FAULTY_FILE; }
  if (!build_xit_rit_mapping(&computation)) {
   free_computation(&computation);
   return ERROR("Could not build xit-rit mapping."), EXIT_FAILURE; }
  successful = compute_computation(&pargs, &computation); }

 else if (pargs.mode_edit) {
  if (!read_computation(pargs.filename, &computation)) { return FAULTY_FILE; }
  successful = edit_computation(&pargs, &computation); }

 else return ERROR("Fatal error -- unknown mode."), EXIT_FAILURE;

 #undef FAULTY_FILE
 free_computation(&computation);
 free_xit_rit_mapping();
 if (successful) LOG("Exiting successfully.");
 return successful ? EXIT_SUCCESS : EXIT_FAILURE; }

